*------------------------------------------------------------------------+
| PR-OAC Electoral System - Simulation and Evaluation                    |
| (C) 2019 by Prof. Werner Antweiler, University of British Columbia     |
+------------------------------------------------------------------------;
options nocenter ls=80;
libname here '.';
%LET PRINT=0;
%LET METHOD=WEBSTER;
%LET SEATS=100;
%LET ITERATIONS=2500;

%MACRO SIMULATE(PARTIES,MODEL,VLIST);
%DO ITER=1 %TO &ITERATIONS;
data election;
  *--------------------------------------------------+
  | generate a randomized election outcome           |
  | draw from Dirichlet distribution                 |
  | with mean mu[i]=alpha[i]/xi                      |
  | xi=sum(alpha[i],i=1..k)                          |
  | and variance sigma[i]^2=mu[i]*(1-mu[i])/(1+xi)   |
  +--------------------------------------------------;
  seats=&SEATS;
  parties=&PARTIES;
  array votes[1:&PARTIES] v1-v&PARTIES;
  array alpha[1:&PARTIES] alpha1-alpha&PARTIES;
  array mu[1:&PARTIES] _temporary_ (&VLIST);
  *call streaminit(12345);
  array noise[0:1] _temporary_ (5 25);
  if &ITERATIONS=1 then chi=0.5;
  else chi=(&ITER-1)/(&ITERATIONS-1);
  xi=(1-chi)*noise[0]+chi*noise[1];
  call symput('XI',put(xi,best10.));
  length seat $4;
  do r=1 to seats;
    seat='R'||put(r,Z3.);
    *--------------------------------------------------+
    | draw {k} random numbers from Gamma distribution  |
    | and then transform into Dirichlet distribution   |
    +--------------------------------------------------;
    do i=1 to parties;
      alpha[i]=rand('GAMMA',xi*mu[i]/100);
    end;
    vsum=sum(of alpha[*]);
    entropy=0;
    do i=1 to parties;
      alpha[i]=alpha[i]/vsum;
      entropy=entropy-log(alpha[i])*alpha[i];
    end;
    *--------------------------------------------------+
    | Determine First-Past-the-Post result and         |
    | turn shares into integer results based on        |
    | 100,000 voters per constituency                  |
    +--------------------------------------------------;
    winner=0; maxvotes=0; allvotes=0;
    do i=1 to parties;
      votes[i]=floor(1E5*alpha[i]+0.5);
      if votes[i]>maxvotes then do;
        winner=i;
        maxvotes=votes[i];
      end;
      allvotes=allvotes+votes[i];
    end;
    represented0=maxvotes;
    output;
  end;
  keep seat v1-v&PARTIES entropy winner represented0 allvotes;
  label seat='Seat/Constituency/Riding Identifier'
    entropy='Entropy Index of Dispersion'
    winner='Party of Winner in FPTP'
    represented0='Voters represented under FPTP'
    allvotes='Sum of votes (needed because of rounding)';
proc sort data=election;
  by seat;
data PR_seats;
  set election end=verylast;
  retain print &PRINT;
  if print then  file 'simulation.txt';
  parties=&PARTIES;
  *--------------------------------------------------+
  | Determine Proportional Representation results    |
  +--------------------------------------------------;
  array votes[1:&PARTIES] v1-v&PARTIES;
  array seats[1:&PARTIES] _temporary_;
  array total[0:&PARTIES] _temporary_;
  if _N_=1 then do;
    total[0]=0;
    do i=1 to parties;
      seats[i]=0;
      total[i]=0;
    end;
  end;
  if print then put seat $4. ': ' allvotes comma7.0 ' :' @;
  do i=1 to parties;
    pct=100*votes[i]/allvotes;
    if print then  put '  ' pct 5.2 @;
    total[i]=total[i]+votes[i];
  end;
  total[0]=total[0]+allvotes;
  seats[winner]=seats[winner]+1;
  if print then do;
    code=substr('ABCDEFGHIJK',winner,1);
    put ' : ' code $1. @;
    N=exp(entropy);
    put ' : ' N 5.3;
  end;
  drop i pct code;
  if verylast then do;
    if print then  do;
      put; put 'First Past the Post';
      do i=1 to parties;
        code=substr('ABCDEFGHIJK',i,1);
        put 'Party ' code $1. ':' @;
        share=100*seats[i]/&SEATS;
        put ' ' seats[i] 3.0 ' (' share 5.2 '%) seats' @;
        share=100*total[i]/total[0];
        put ' ' share 5.2 '% votes';
      end;
    end;
    *--------------------------------------------------+
    | Gallagher Index of Disproportionality            |
    +--------------------------------------------------;
    gallagher=0;
    do i=1 to parties;
       s_share=100*seats[i]/&SEATS;
       v_share=100*total[i]/total[0];
       gallagher=gallagher+(s_share-v_share)**2;
    end;
    gallagher=sqrt(gallagher/2);
    label gallagher='Gallagher Index of Disproportionality';
    if print then do;
      put 'Gallagher Index: ' gallagher 8.3;
    end;
    *--------------------------------------------------+
    | Allocation of seats based on d_Hondt             |
    | or by Webster/Sainte-Lagu method                |
    +--------------------------------------------------;
    array x[&PARTIES] _temporary_;
    array z[&PARTIES] _temporary_;
    array s[&PARTIES] PR_seats1-PR_seats&PARTIES;
    do i=1 to parties;
      x[i]=total[i];
      s[i]=0;
      z[i]=x[i];
    end;
    do j=1 to &SEATS;
      maxz=0; maxi=0;
      do i=1 to parties;
        if z[i]>maxz then do; maxz=z[i]; maxi=i; end;
      end;
      s[maxi]=s[maxi]+1;
      if "&METHOD"='DHONDT' then z[maxi]=x[maxi]/(1+s[maxi]);
      else if "&METHOD"='WEBSTER' then z[maxi]=x[maxi]/(1+2*s[maxi]);
      else put 'unspecified method';
    end;
    length majority $1;
    majority='X';
    threshold=ceil(&SEATS/2);
    do i=1 to &PARTIES;
      if s[i]>=threshold then majority=substr('ABCDEFGHIJK',i,1);
    end;
    call symput('MAJORITY',majority);
    if print then do;
      put; put 'Proportional Representation';
      do i=1 to parties;
        code=substr('ABCDEFGHIJK',i,1);
        put 'Party ' code $1. ':' @;
        share=100*s[i]/&SEATS;
        put ' ' s[i] 3.0 ' (' share 5.2 '%) seats' @;
        share=100*total[i]/total[0];
        put ' ' share 5.2 '% votes';
      end;
      if majority='X' then put 'Plurality but no majority';
      else put 'Party ' majority $1. ' has majority of seats';
    end;
    output;
  end;
  keep majority PR_seats1-PR_seats&PARTIES gallagher;
data cplex;
  set election end=verylast;
  file 'election1.dat';
  *--------------------------------------------------+
  | Generate input file for CPLEX/OPL software       |
  | Standard version for single-seat PR-OAC          |
  | and dual-seat PR-OAC without majority            |
  +--------------------------------------------------;
  length code $1;
  if _N_=1 then do;
    put '/*----------------------------------------------------';
    put '  CPLEX OPL Maximization Problem';
    put '  Simulated Election Data';
    put '  Werner Antweiler, University of British Columbia';
    put '----------------------------------------------------*/';
    put 'Parties = {' @;
    do i=1 to &PARTIES;
      code=substr('ABCDEFGHIJKLM',i,1);
      put '"' code $1. '"' @;
      if i<&PARTIES then put ',' @;
    end;
    put '};';
    put 'Ridings = {' @;
    do i=1 to &SEATS;
      put '"R' i Z3.0 '"' @;
      if i=&SEATS then put '};';
      else if mod(i-1,10)=0 then put ',';
      else put ',' @;
    end;
    set PR_seats;
    array s[&PARTIES] PR_seats1-PR_seats&PARTIES;
    put 'Seats = [' @;
    do i=1 to &PARTIES;
      put s[i] +(-1) @;
      if i<&PARTIES then put ',' @;
      else put '];';
    end;
    put 'Nodes = {';
  end;
  array v[&PARTIES] v1-v&PARTIES;
  do i=1 to &PARTIES;
    code=substr('ABCDEFGHIJKLM',i,1);
    put '  <"' code $1. '","' seat $4. '",'  v[i] +(-1) '>' @;
    if i<&PARTIES then put ',';
  end;
  if verylast then put '};';
  else put ',';
  drop _ALL_; delete;
data cplex;
  set election end=verylast;
  file 'election2.dat';
  *--------------------------------------------------+
  | Generate input file for CPLEX/OPL software       |
  | special case for dual-seat PR-OAC with majority  |
  +--------------------------------------------------;
  length code $1;
  length majority $1;
  retain majority;
  if _N_=1 then do;
    majority=symget('MAJORITY');
    put '/*----------------------------------------------------';
    put '   CPLEX OPL Maximization Problem';
    put '   Simulated Election Data';
    put '   Werner Antweiler, University of British Columbia';
    put '----------------------------------------------------*/';
    put 'Parties = {' @;
    do i=1 to &PARTIES;
      code=substr('ABCDEFGHIJKLM',i,1);
      put '"' code $1. '"' @;
      if i<&PARTIES then put ',' @;
    end;
    put '};';
    put 'Ridings = {' @;
    do i=1 to &SEATS;      
      put '"R' i Z3.0 '"' @;
      if i=&SEATS then put '};';
      else if mod(i-1,10)=0 then put ',';
      else put ',' @;
    end;
    set PR_seats;
    array s[&PARTIES] PR_seats1-PR_seats&PARTIES;
    put 'Seats = [' @;
    do i=1 to &PARTIES;
      code=substr('ABCDEFGHIJKLM',i,1);
      if majority='X' then put s[i] +(-1) @;
      else do;
        if majority^=code then needed=2*s[i];
        else needed=2*s[i]-&SEATS;
        put needed +(-1) @;
      end;
      if i<&PARTIES then put ',' @;
      else put '];';
    end;
    put 'Nodes = {';
  end;
  array v[&PARTIES] v1-v&PARTIES;
  do i=1 to &PARTIES;
    code=substr('ABCDEFGHIJKLM',i,1);
    put '  <"' code $1. '","' seat $4. '",' @;
    *--------------------------------------------------+
    | a majority party gets one seat in each riding    |
    | and remaining seats are based on zero extra      |
    | representation                                   |
    +--------------------------------------------------;    
    if code=majority then put '0' @;
    else put v[i] +(-1) @;
    put '>' @;
    if i<&PARTIES then put ',';
  end;
  if verylast then put '};';
  else put ',';
  drop _ALL_; delete;
run;

data null;
  *--------------------------------------------------+
  | run the simulation with CPLEX                    |
  +--------------------------------------------------;
  call system("/usr/local/bin/cplex-opl election1.mod election1.dat </dev/null >election1.log");
  call system("/bin/gawk -f election.awk election1.log >election1.opt");
  *--------------------------------------------------+
  | for dual-seat PR-OAC, run two-seat model for case|
  | without majority, and single-seat truncated case |
  | when one party has a majority and gets at least  |
  | one seat in each riding                          |
  +--------------------------------------------------;
  if symget('MAJORITY')='X' then
    call system("/usr/local/bin/cplex-opl election2.mod election2.dat </dev/null >election2.log");
  else
    call system("/usr/local/bin/cplex-opl election1.mod election2.dat </dev/null >election2.log");
  call system("/bin/gawk -f election.awk election2.log >election2.opt");
run;

data results1;
  *--------------------------------------------------+
  | Read results from single-seat PR-OAC simulation  |
  +--------------------------------------------------;
  infile "election1.opt" dsd delimiter=',';
  if _N_=1 then do;
    input optimum;
    retain optimum;
  end;
  else do;
    length seat $4;
    length assigned $1;
    input seat $ assigned $ represented1;
    output;
  end;
  keep seat assigned represented1;
proc sort data=results1;
  by seat;
data results2;
  *--------------------------------------------------+
  | Read results from dual-set PR-OAC simulation     |
  +--------------------------------------------------;
  infile "election2.opt" dsd delimiter=',';
  if _N_=1 then do;
    input optimum;
    retain optimum;
  end;
  else do;
    length seat $4;
    length party $1;
    input seat $ party $ votes;
    output;
  end;
  keep seat party votes;
proc sort data=results2;
  by seat descending votes;
data results2;
  set results2;
  by seat descending votes;
  retain represented2 rank;
  length junior senior $1;
  if first.seat then do;
    represented2=0; rank=0;
  end;
  represented2=represented2+votes;
  rank=rank+1;
  if first.seat=last.seat then do; 
    * majority case ;
    senior=symget('MAJORITY');
    junior=party;
  end;
  else do;
    if rank=1 then senior=party;
    else if rank=2 then junior=party;
    else put 'ERROR: rank>2';
  end;
  if last.seat then output;
  keep seat junior senior represented2;
data election;
  *--------------------------------------------------+
  | Merge results and analyze                        |
  +--------------------------------------------------;
  merge election results1 results2;
  by seat;
  *--------------------------------------------------+
  | adjust total representation for dual-seat PR-OAC |
  | when one party has a majority of seats           |
  +--------------------------------------------------;
  length majority $1;
  majority=symget('MAJORITY');
  array votes[1:&PARTIES] v1-v&PARTIES;
  if majority^='X' then do;
    do i=1 to &PARTIES;
      code=substr('ABCDEFGHIJKLM',i,1);
      if code=majority   /* majority party votes */
      and junior^=senior /* do not double count */
      then represented2=represented2+votes[i];
    end;
  end;
data summary;
  set election end=verylast;
  by seat;
  retain rep0 rep1 rep2 total 0;
  rep0=rep0+represented0;
  rep1=rep1+represented1;
  rep2=rep2+represented2;
  total=total+allvotes;
  *put seat $4.
    ' ' represented0 comma7.0 
    ' ' represented1 comma7.0
    ' ' represented2 comma7.0;
  if verylast then do;
    set PR_seats;
    length model $1;
    model="&MODEL";
    seats=&SEATS;
    parties=&PARTIES;
    noise=input(symget('XI'),10.);
    rep0=100*rep0/total;
    rep1=100*rep1/total;
    rep2=100*rep2/total;
    iteration=&ITER;
    output;
  end;
  keep iteration noise seats parties model noise
    rep0 rep1 rep2 gallagher majority;
proc append base=simulation data=summary;
run;
%END;
%MEND;

*------------------------------------------------------------------------+
| SIMULATE(PARTIES,MODEL,VOTESHARELIST)                                  |
+------------------------------------------------------------------------;
options nosource nonotes;
%SIMULATE(2,A,60 40);
%SIMULATE(2,B,52 48);
%SIMULATE(3,A,55 35 10);
%SIMULATE(3,B,45 35 20);
%SIMULATE(3,C,40 35 25);
%SIMULATE(5,A,50 20 15 10  5);
%SIMULATE(5,B,40 25 20 10  5);
%SIMULATE(5,C,30 25 20 15 10);
options source notes;

proc sort data=simulation out=here.simulation;
  by parties model iteration;
proc print data=here.simulation;
  id parties model iteration;
  var rep0 rep1 rep2;
run;
endsas;

